import abc
from typing import NamedTuple, List

import pytest

from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import QueueReporterIF, ExecutionManagerIF, QueueReport
from workspaces.capability.services.queue_manager import CapabilityQueueManager


class FakeQueueReporterExecutionManager(QueueReporterIF, ExecutionManagerIF):
    """
    A testing mock that pretends to be both the queue reporter and execution manager.
    Our notions of both are pretty limited, basically moving things from one list to
    another so that we can get proper reports, but otherwise not doing anything real.
    """

    submitted = []
    executing = []
    completed = []
    allowed = 3

    def report(self) -> List[QueueReport]:
        return [QueueReport("fake", len(self.submitted), len(self.executing), self.allowed)]

    def waiting_executions(self, queue: str, count: int) -> List[CapabilityExecutionIF]:
        return self.submitted[:count]

    def submit(self, execution: CapabilityExecutionIF):
        self.submitted.append(execution)

    def execute(self, execution: CapabilityExecutionIF):
        self.submitted.remove(execution)
        self.executing.append(execution)

    def complete(self, execution: CapabilityExecutionIF):
        self.executing.remove(execution)
        self.completed.append(execution)


@pytest.mark.skip("Broken due to queue/messenger rework")
def test_capability_queue_manager():
    # let's start by putting five things into the execution manager
    test_reporter_executor = FakeQueueReporterExecutionManager()
    for i in range(1, 6):
        test_reporter_executor.submit(i)

    # now let's make our queue manager and see what the report tells us
    queue = CapabilityQueueManager(test_reporter_executor, test_reporter_executor)

    # we should now have 3 executing and 2 submitted
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].submitted == 2

    # now let's finish a couple of them
    test_reporter_executor.complete(1)
    queue.execution_complete()
    test_reporter_executor.complete(2)
    queue.execution_complete()

    # we should now have 3 executing and 0 submitted
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].submitted == 0

    # let's complete a couple more, this time no notification in between
    test_reporter_executor.complete(3)
    test_reporter_executor.complete(4)
    queue.execution_complete()

    # we should now have 1 executing and 0 submitted
    assert test_reporter_executor.report()[0].executing == 1
    assert test_reporter_executor.report()[0].submitted == 0

    # let's add four more
    for i in range(6, 10):
        test_reporter_executor.submit(i)
    queue.execution_requested()

    # now check again and we should have 3 executing and 2 submitted again
    assert test_reporter_executor.report()[0].executing == 3
    assert test_reporter_executor.report()[0].submitted == 2
