""" This is the execution manager. """

import logging
from typing import Dict, List, Optional

from messaging.messenger import MessageSender
import transaction
from messaging.router import Router, on_message

from workspaces.capability.enums import (
    CapabilityRequestState,
    CapabilityStepType,
    ExecutionState,
)
from workspaces.capability.helpers import CapabilitySequence
from workspaces.capability.message_architect import CapabilityMessageArchitect
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    ExecutionManagerIF,
)
from workspaces.capability.services.queue_manager import CapabilityQueueManager
from workspaces.system.schema import AbstractFile
from workspaces.workflow.services.interfaces import WorkflowServiceIF

# pylint: disable=E0401, W1203


logger = logging.getLogger(__name__)


class ExecutionManager(ExecutionManagerIF):
    """
    Class that manages the capability execution system. It's duties include:
    - Creating new executions
    - Managing existing execution state
    - Handling capability events and updating the state of the system accordingly
    """

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        router: Router,
        messenger: MessageSender,
    ):
        self.capability_info = capability_info
        self.workflow_service = workflow_service
        self.messenger = messenger
        self.queue_manager = CapabilityQueueManager(self.capability_info, self)
        self.router = router
        self.router.register(self)

    def run_capability(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Run a capability with the requested settings

        :param execution: Capability execution
        :return: Execution of the request
        """
        request = execution.capability_request
        request.state = CapabilityRequestState.Submitted.name
        self.capability_info.save_entity(request)
        return self.execute_or_queue(execution)

    def execute_or_queue(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Execute this execution, or else put it in the queue (if it needs to run a PrepareAndRunWorkflow)

        :param execution:
        :return:
        """
        # if this is PrepareAndRunWorkflow, we enter the queue by going into the Waiting state instead
        if execution.current_step.step_type == CapabilityStepType.PrepareAndRunWorkflow:
            logger.info("Sticking PrepareAndRunWorkflow into the queue")
            execution.state = ExecutionState.Waiting.name
            self.capability_info.save_execution(execution)
            transaction.commit()
            self.queue_manager.execution_requested()
        else:
            logger.info("Proceding with execution")
            # otherwise, we proceed with execution
            self.execute(execution)
        return execution

    def execute(self, execution: CapabilityExecutionIF):
        """
        Communicate with workflow service and send a request to run a workflow with given settings
        """
        logger.info("Asking step %s to execute in execution %s", execution.current_step_idx, execution.id)
        execution.current_step.execute(self, execution)

    @on_message(service="workflow", type="workflow-complete")
    @on_message(service="workflow", type="workflow-failed")
    def update_capability_queue_manager(self, **msg):
        self.queue_manager.execution_complete()

    @on_message(service="workflow", type="workflow-executing")
    @on_message(service="workflow", type="workflow-complete")
    @on_message(service="workflow", type="workflow-failed")
    @on_message(service="workflow", type="qa-pass")
    @on_message(service="workflow", type="qa-fail")
    def advance_waiting_execution(self, **message: Dict):
        """
        Advance execution that is waiting to its next step

        :param message: Workflow-execution message
        """
        logger.info(f"Received workflow message: {message}")
        logger.info("Attempting to advance an execution based on this message...")
        workflow_request_id = self._get_workflow_request_id_from_msg(**message)
        if workflow_request_id is not None:
            execution = self.capability_info.lookup_execution_by_workflow_request_id(workflow_request_id)
            if execution is not None:
                self._advance_execution(execution)
            else:
                logger.error(
                    f"No capability execution found corresponding to workflow request #{workflow_request_id}; no executions advanced"
                )
        else:
            logger.error(
                "Workflow executing message does not have a workflow request associated with it; no executions advanced"
            )

    @on_message(service="workflow", type="workflow-failed")
    def on_workflow_failed(self, **message: Dict):
        """
        Announce and record workflow failure

        :param message: the failure message
        """
        logger.info(f"RECEIVED WORKFLOW-FAILED: {message}")
        workflow_request_id = self._get_workflow_request_id_from_msg(**message)
        execution = self.capability_info.lookup_execution_by_workflow_request_id(workflow_request_id)

        if execution:
            execution.state = ExecutionState.Failed.name

            execution_failed_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_failed")
            self.messenger.send_message(**execution_failed_msg)
            self.capability_info.save_execution(execution)
        else:
            logger.error(f"No capability execution found for workflow id {workflow_request_id}; no executions failed")

    @on_message(message="Ingestion complete")
    def on_ingestion_complete(self, **message: Dict):
        """
        Announce and record a completed ingestion event.

        :param message: the ingestion-complete message
        :return:
        """
        logger.info(f"This is a ingestion-complete message: {message}")

    @on_message(service="capability", type="execution-failed")
    def on_execution_fail(self, **message: Dict):
        """
        Announce and record execution failure
        # TODO: Determine if this is made redundant by on_workflow_failed()

        :param message: the failure message
        """
        logger.info(f"RECEIVED EXECUTION-FAILED: {message}")
        subject = message["subject"]
        execution = self.capability_info.lookup_execution_by_workflow_request_id(subject["current_workflow_request_id"])
        # Record failure in databse
        execution.state = ExecutionState.Failed.name
        self.capability_info.save_execution(execution)

    def submit_workflow_request(
        self,
        execution_id: int,
        workflow_name: str,
        workflow_args: dict,
        files: List[AbstractFile],
    ):
        """
        Submit a workflow request to the workflow service

        :param execution_id:
        :param workflow_name:
        :param workflow_args:
        :param files:
        """
        logger.info("Submitting workflow request")
        transfer_files_string = self._get_transfer_files_property(files)

        args_list = {**workflow_args, "files_to_transfer": transfer_files_string}

        logger.info(
            "creating workflow request with name %s, args %s (and arg type %s)",
            workflow_name,
            args_list,
            type(args_list),
        )
        workflow_request = self.workflow_service.create_workflow_request(workflow=workflow_name, argument=args_list)

        logger.info(
            "My workflow_request is %s, with type %s and workflow request ID %s",
            workflow_request,
            type(workflow_request),
            workflow_request.workflow_request_id,
        )

        for file in files:
            self.workflow_service.attach_file_to_request(workflow_request, file.filename, file.content)
            logger.info(f"Added file {file.filename} to workflow request #{workflow_request.workflow_request_id}")

        execution = self.capability_info.lookup_execution(execution_id)
        execution.state = ExecutionState.ExecutingWorkflow.name
        execution.update_execution(workflow_request.workflow_request_id)
        self.capability_info.save_execution(execution)
        self.workflow_service.execute(workflow_request)

    @staticmethod
    def _get_workflow_request_id_from_msg(**message: Dict) -> Optional[int]:
        """
        Extract workflow request ID from workflow message

        :param message: Workflow message
        :return: Workflow request ID if found in message; else None
        """
        workflow_request_id = message.get("subject").get("workflow_request_id")
        if workflow_request_id is not None:
            # Cast from str to int
            workflow_request_id = int(workflow_request_id)
            return workflow_request_id
        else:
            return None

    def _advance_execution(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Perform necessary steps to advance an execution to its next step

        :param execution: Execution to advance
        :return: Execution with updated metadata
        """
        step_sequence = CapabilitySequence.from_str(execution.steps)

        if execution.on_last_step():
            self._complete_execution(execution)
        else:
            execution.current_step_idx += 1
            self.execute_or_queue(execution)

        self.capability_info.save_execution(execution)
        return execution

    def _complete_execution(self, execution: CapabilityExecutionIF) -> CapabilityExecutionIF:
        """
        Perform necessary steps to declare a capability execution complete

        :param execution: Execution to complete
        :return: Execution with updated metadata
        """
        execution.state = ExecutionState.Complete.name

        execution_complete_msg = CapabilityMessageArchitect(execution=execution).compose_message("execution_complete")
        self.messenger.send_message(**execution_complete_msg)

        return execution

    @staticmethod
    def _get_transfer_files_property(files: List[AbstractFile]):
        separator = ", ./"
        return separator + separator.join(file.filename for file in files)
