import glob
import os
from pathlib import Path

import logging
from pycapo import CapoConfig
from immutable_views import ListView
from workspaces.system.schema import AbstractFile

from workspaces.system.enums import RemoteWorkflows, RemoteProcessingSites
from workspaces.system.views import ViewWorkflowCapoParameters
from workspaces.workflow.schema_interfaces import WorkflowRequestIF

logger = logging.getLogger(__name__)


class CapoInjector:
    """
    Class for determining and creating a subspace, or subset, of Capo properties specific
    to functionality of a remote workflow
    """

    def __init__(self, wf_request: WorkflowRequestIF):
        self.request = wf_request
        self.profile = self.set_profile()
        self.views = ViewWorkflowCapoParameters(wf_request)
        self.dir_path = Path(self.request.results_dir)

    def set_profile(self):
        current = CapoConfig().profile
        logger.info(f"current profile: {current}")

        if RemoteProcessingSites.NAASC.value in self.request.argument:
            system = current.split("-")[1]
            remote_profile = "naasc-" + system
            logger.info(f"profile for request: {remote_profile}")
            return remote_profile

        return current

    def is_remote_workflow(self) -> bool:
        """
        Determine if the provided request is for a workflow that can run remotely

        :return: boolean representing if this workflow request is for a remote executing workflow
        """
        parameters = self.request.argument
        remote_types = RemoteProcessingSites.values()
        return any([key in remote_types for key in parameters]) and RemoteWorkflows.is_remote(
            wf_name=self.request.workflow_name
        )

    def identify_subspace(self) -> ListView:
        """
        Determine specific capo requirements for a remote workflow

        :return: Immutable list containing required capo properties
        """
        return self.views.construct_view()

    def create_subspace(self) -> list:
        """
        Pull Capo settings for required properties

        :return: Immutable list containing capo property strings
        """
        requirements = self.identify_subspace()

        subspace = []

        for item in requirements:
            setting = CapoConfig(profile=self.profile).get(item)
            subspace.append(item + " = " + setting)

        return subspace

    def write_subspace(self) -> AbstractFile:
        """
        Write workflow capo subspace to new file.

        :return: Abstract file containing the workflow specific capo properties required to run remotely.
        """
        filename = self.profile + ".properties"
        content = "\n".join(self.create_subspace())

        return AbstractFile(filename=filename, content=str(content).encode())

    def subspace_to_directory(self):
        """
        Write new subspace file to request's results_dir

        :return: None
        """
        subspace = self.write_subspace()

        path = self.dir_path / subspace.filename
        path.write_bytes(subspace.content)
        logger.info(f"Writing capo subspace file to {self.dir_path.__str__()}")

    def clear_subspace(self):
        logger.info(f"Clearing capo subspace file from {self.dir_path.__str__()}...")
        for file in os.listdir(self.dir_path):
            if file.endswith(".properties"):
                Path.unlink(self.dir_path / file)

        check = glob.glob("*.properties")
        if check is None:
            logger.info("Capo subspace cleared successfully.")
